#include "adb.h"
#include "ps2.h"
#include "usart.h"

__IO struct adb adb;
__IO uint8_t adbkey[kADBBuffSize];
__IO uint8_t adbstart = 0;
__IO uint8_t adbend = 0;

extern uint8_t ps2ledstate;
extern uint8_t modifierkeys;

void ADBTick_Handler(void) {
	GPIO_InitTypeDef GPIO_InitStructure;
	uint16_t tmpdata;
	uint8_t tmpticks;

	switch(adb.state) {
	case kADBStateIdle:
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 0) {
			adb.state = kADBStateAttn;
			adb.curticks = 0;
		}
		break;
	case kADBStateAttn:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 1) {
			// If the attention signal was for less than 800us,
			// it wasn't long enough to be an attn signal
			if(adb.curticks < 79) {
				adb.state = kADBStateIdle;
				break;
			}

			adb.state = kADBStateSync;
			adb.curticks = 0;
		}
		break;
	case kADBStateSync:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 0) {
			adb.state = kADBStateReadCmdBitLow;
			adb.curticks = 0;
			adb.bit = 0;
			adb.cmd = 0;
		}
		break;
	case kADBStateReadCmdBitLow:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 1) {
			adb.state = kADBStateReadCmdBitHigh;
			if(adb.curticks < 5) {
				adb.cmd |= 1<<(7-adb.bit);
			}
			adb.curticks = 0;
		}
		break;
	case kADBStateReadCmdBitHigh:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 0) {
			adb.state = kADBStateReadCmdBitLow;
			if(adb.bit >= 7) {
				adb.state = kADBStateStopLow;
			}
			adb.bit++;
			adb.curticks = 0;
		}
		break;
	case kADBStateStopLow:
		adb.curticks++;
		// If the currently addressed device does not have data to send,
		// see if any of our other devices do, and issue service req.
		if(adb.senddata && !(adb.senddata & (1<<((adb.cmd & 0xF0) >> 4)))) {
			adb.state = kADBStateSignalRequest;
			adb.curticks = 0;
			GPIO_InitStructure.GPIO_Pin = kADBPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
			GPIO_Init(GPIOB, &GPIO_InitStructure);
			GPIO_ResetBits(GPIOB, kADBPin);
		}
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 1) {
			uint8_t devid = (adb.cmd & 0xF0) >> 4;
			// Do we need a stop-to-start time state?
			adb.state = kADBStateIdle;
			// If the cmd is addressed to any of our devices
			if(adb.activedevs & (1<<devid)) {
				adb.state = kADBStateStopStart;
			}
			adb.curticks = 0;
			adb.bit = 0;
		}
		break;
	case kADBStateSignalRequest:
		adb.curticks++;

		if(adb.curticks > 14) {
			adb.state = kADBStateIdle;
			GPIO_InitStructure.GPIO_Pin = kADBPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
			GPIO_Init(GPIOB, &GPIO_InitStructure);
		}
		break;
	case kADBStateStopStart:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 0) {
			adb.state = kADBStateIdle;
			switch(adb.cmd & 0x0C) {
			case 0x08:
				// Listen
				adb.state = kADBStateReadDataBitLow;
				adb.curticks = 0;
				adb.bit = 0;
				switch(adb.cmd & 0x03) {
				case 0: adb.device[(adb.cmd&0xF0) >> 4].reg0 = 0; break;
				case 1: adb.device[(adb.cmd&0xF0) >> 4].reg1 = 0; break;
				case 2: adb.device[(adb.cmd&0xF0) >> 4].reg2 = 0; break;
				case 3: adb.device[(adb.cmd&0xF0) >> 4].reg3 = 0; break;
				}
				break;
			} 
		} else {
			uint8_t devid = (adb.cmd & 0xF0) >> 4;
			uint8_t adbleds = 0;
			if(adb.curticks > 26) adb.state = kADBStateIdle;

			if((adb.curticks > 19) && ((adb.cmd & 0x0C) == 0x0C) && (adb.senddata & (1<<devid))) {
				// Talk

				// Special case keyboard...
				switch((adb.cmd&0xF0) >> 4) {
				case 2:
					switch(adb.cmd&0x03) {
					case 0:
						if(adbstart == adbend) {
							adb.state = kADBStateIdle;
							break;
						} else {
							adb.device[2].reg0 = adbkey[adbstart] << 8;
							adbstart = (adbstart+1) % kADBBuffSize;
							if(adbstart != adbend) {
								adb.device[2].reg0 |= adbkey[adbstart];
								adbstart = (adbstart+1) % kADBBuffSize;
							} else {
								adb.device[2].reg0 |= 0x7F;
							}
						}
						break;
					case 2:
						adbleds = 0xFF;
						if(!(ps2ledstate & kPS2LEDCaps)) adbleds &= ~2;
						if(!(ps2ledstate & kPS2LEDScroll)) adbleds &= ~4;
						if(!(ps2ledstate & kPS2LEDNum)) adbleds &= ~1;
						adb.device[2].reg2 = (modifierkeys << 8) | adbleds;
						break;
					default: break;
					}
				default:
					break;
				}
				adb.state = kADBStateWriteStartBitLow;
				adb.curticks = 0;
				adb.bit = 0;
				GPIO_InitStructure.GPIO_Pin = kADBPin;
				GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
				GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
				GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
				GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
				GPIO_Init(GPIOB, &GPIO_InitStructure);
				GPIO_ResetBits(GPIOB, kADBPin);
			}
		}
		break;
	case kADBStateReadDataBitLow:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 1) {
			adb.state = kADBStateReadDataBitHigh;
			if(adb.curticks < 4) {
				switch(adb.cmd & 0x03) {
				case 0: adb.device[(adb.cmd&0xF0) >> 4].reg0 = 1<<(15-adb.bit); break;
				case 1: adb.device[(adb.cmd&0xF0) >> 4].reg1 = 1<<(15-adb.bit); break;
				case 2: adb.device[(adb.cmd&0xF0) >> 4].reg2 = 1<<(15-adb.bit); break;
				case 3: adb.device[(adb.cmd&0xF0) >> 4].reg3 = 1<<(15-adb.bit); break;
				}
			}
			adb.curticks = 0;
		}
		break;
	case kADBStateReadDataBitHigh:
		adb.curticks++;
		if(GPIO_ReadInputDataBit(GPIOB, kADBPin) == 0) {
			adb.state = kADBStateReadDataBitLow;
			if(adb.bit >= 7) {
				adb.state = kADBStateStopLow;
			}
			adb.bit++;
			adb.curticks = 0;
			adb.cmd = 0; // Zero the command so kADBStateStopLow doesn't send us back here.
		}
		break;
	case kADBStateWriteStartBitLow:
		adb.curticks++;
		if(adb.curticks >= 4) {
			adb.state = kADBStateWriteStartBitHigh;
			adb.curticks = 0;
			GPIO_SetBits(GPIOB, kADBPin);
		}
		break;
	case kADBStateWriteStartBitHigh:
		adb.curticks++;
		if(adb.curticks >= 7) {
			adb.state = kADBStateWriteDataBitLow;
			adb.curticks = 0;
			GPIO_ResetBits(GPIOB, kADBPin);
		}
		break;
	case kADBStateWriteDataBitLow:
		adb.curticks++;
		switch(adb.cmd & 0x03) {
		case 0: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg0; break;
		case 1: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg1; break;
		case 2: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg2; break;
		case 3: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg3; break;
		}
		if(tmpdata & (1<<(15-adb.bit))) {
			tmpticks = 3;
		} else {
			tmpticks = 7;
		}
		if(adb.curticks >= tmpticks) {
			adb.state = kADBStateWriteDataBitHigh;
			adb.curticks = 0;
			GPIO_SetBits(GPIOB, kADBPin);
		}
		break;
	case kADBStateWriteDataBitHigh:
		adb.curticks++;
		switch(adb.cmd & 0x03) {
		case 0: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg0; break;
		case 1: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg1; break;
		case 2: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg2; break;
		case 3: tmpdata = adb.device[(adb.cmd&0xF0) >> 4].reg3; break;
		}
		if(tmpdata & (1<<(15-adb.bit))) {
			tmpticks = 7;
		} else {
			tmpticks = 3;
		}
		if(adb.curticks >= tmpticks) {
			if(adb.bit > 15) {
				adb.state = kADBStateWriteStop;
			} else {
				adb.state = kADBStateWriteDataBitLow;
				GPIO_ResetBits(GPIOB, kADBPin);
				adb.bit++;
			}
			adb.curticks = 0;
		}
		break;
	case kADBStateWriteStop:
		adb.curticks++;
		if(adb.curticks > 7) {
			uint8_t devid = (adb.cmd & 0xF0) >> 4;
			adb.state = kADBStateIdle;
			GPIO_InitStructure.GPIO_Pin = kADBPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
			GPIO_Init(GPIOB, &GPIO_InitStructure);
			if((devid != 2) || ((devid == 2) && (adbstart == adbend))) {
				adb.senddata &= ~(1<<devid);
			}
			adb.curticks = 0;
		}
		break;
	}
}

void ADBInit(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	int i;

	GPIO_InitStructure.GPIO_Pin = kADBPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOB, &GPIO_InitStructure);

	GPIO_InitStructure.GPIO_Pin = kADBPowerPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOB, &GPIO_InitStructure);
	GPIO_SetBits(GPIOB, kADBPowerPin);

	adb.state = kADBStateIdle;
	adb.bit = 0;
	adb.curticks = 0;
	adb.cmd = 0;
	adb.senddata = 0;
	adb.activedevs = 0;
	for(i = 0; i < 16; i++) {
		adb.device[i].reg0 = 0;
		adb.device[i].reg1 = 0;
		adb.device[i].reg2 = 0;
		adb.device[i].reg3 = 0;
		adb.device[i].devid = 0;
	}
	adb.device[2].devid = 0x20;
	adb.device[2].reg2 = 0xFF00;
	adb.device[2].reg3 = 0x6203;
	adb.activedevs |= 1<<(0x20>>4);
	adb.device[3].devid = MYDEVID;
	adb.device[3].reg3 = 0x63AA;
	adb.activedevs |= 1<<(MYDEVID>>4);
}

void ADBSendMouse(uint8_t ps2status, uint8_t ps2xmove, uint8_t ps2ymove)
{
	// If we have an outstanding packet, drop this one
	if(adb.senddata & (1<<3)) {
		return;
	}

	adb.device[3].reg0 = 0;
	if(!(ps2status & 0x01)) adb.device[3].reg0 |= 0x8000;
	if(ps2ymove) {
		adb.device[3].reg0 |= (~ps2status & 0x20) << 9;
		adb.device[3].reg0 |= (((~ps2ymove)+1) & 0x3F) << 8;
	}
	adb.device[3].reg0 |= 0x80; // always 1
	if(ps2xmove) {
		adb.device[3].reg0 |= (ps2status & 0x10) << 2;
		adb.device[3].reg0 |= (ps2xmove & 0x3F);
	}
	adb.senddata |= (1<<3);
}
