#include "stm32f0xx_conf.h"
#include "ps2.h"
#include "usart.h"
#include "adb.h"

#define kBuffSize  32
extern __IO struct adb adb;
extern __IO uint8_t adbkey[kBuffSize];
extern __IO uint8_t adbstart;
extern __IO uint8_t adbend;

uint8_t powerkeydown = 0;

#define kModCmd      1
#define kModOpt      2
#define kModShift    4
#define kModControl  8
#define kModReset   16
#define kModCaps    32
#define kModDelete  64
uint8_t modifierkeys = 0xFF;

void USART2_Callback(uint8_t c);
void HandleMacCommand();
void SendMacResponse(uint8_t respbyte);
uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended);
uint8_t PS2ToADBScancode(uint8_t code, uint8_t keyup, uint8_t extended);

__IO unsigned long tenustimer;
static __IO uint8_t mackey[kBuffSize];
static __IO uint8_t keyindex;
static __IO uint8_t key2index;
static __IO uint8_t keypad;

static __IO uint16_t mousex;
static __IO uint16_t mousey;
static __IO uint16_t mousebits;

#define kPollTimeout 1000

#define kMacKbdClockPin             GPIO_Pin_7
#define kMacKbdDataPin              GPIO_Pin_6

#define kMSEX1 GPIO_Pin_2
#define kMSEX2 GPIO_Pin_3
#define kMSEY1 GPIO_Pin_5
#define kMSEY2 GPIO_Pin_4
#define kMSESW GPIO_Pin_7

#define kMacKbdStateIdle            0
#define kMacKbdStateReadSettle      1
#define kMacKbdStateReading         2
#define kMacKbdStateWriteSettle     3
#define kMacKbdStateWriting         4
#define kMacKbdStateInquiryWait     5
#define kMacKbdStateTest            6
#define kMacKbdStateInstantWait     7

struct mackbd {
	uint8_t state;
	uint8_t bits;
	uint8_t clockstate;
	uint32_t delay;
	uint8_t cmd;
	uint8_t response;
};
static __IO struct mackbd mackbd;

void handle_mouseaxis(__IO uint16_t *movement, uint32_t pinone, uint32_t pintwo) {
	static uint8_t counter = 100;

	if(*movement & 0x00FF) {
		uint8_t movecount = *movement & 0x00FF;

		if(counter--) return;
		counter = 100;

		*movement = (*movement & 0xFF00) | (movecount-1);
		//if((*movement & 0x00FF) < 2) return;

		if(!(*movement & 0x0100)) {
			if(mousebits & pinone) {
				GPIO_ResetBits(GPIOC, pintwo);
				GPIO_ResetBits(GPIOC, pinone);
				mousebits &= ~(pinone | pintwo);
			} else {
				GPIO_SetBits(GPIOC, pintwo);
				GPIO_SetBits(GPIOC, pinone);
				mousebits |= pinone | pintwo;
			}
		} else {
			if(mousebits & pinone) {
				GPIO_SetBits(GPIOC, pintwo);
				mousebits |= pintwo;
				GPIO_ResetBits(GPIOC, pinone);
				mousebits &= ~pinone;
			} else {
				GPIO_ResetBits(GPIOC, pintwo);
				mousebits &= ~pintwo;
				GPIO_SetBits(GPIOC, pinone);
				mousebits |= pinone;
			}
		}
	}
	return;
}

void SysTick_Handler(void) {
	GPIO_InitTypeDef GPIO_InitStructure;
	static uint8_t respbit;
	tenustimer++;

	ADBTick_Handler();

	handle_mouseaxis(&mousex, kMSEX1, kMSEX2);
	handle_mouseaxis(&mousey, kMSEY1, kMSEY2);
	switch(mackbd.state) {
	case kMacKbdStateIdle:
		if(GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) == 0) {
			mackbd.state = kMacKbdStateReadSettle;
			mackbd.delay = 18;
			mackbd.bits = 0;
			mackbd.cmd = 0;
			mackbd.clockstate = 3;
		}
		break;
	case kMacKbdStateReadSettle:
		if(--mackbd.delay == 0) {
			mackbd.state = kMacKbdStateReading;
		}
		break;
	case kMacKbdStateReading:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.delay = 8;
				mackbd.clockstate = 4;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					// Done clocking in all the bits
					mackbd.state = kMacKbdStateIdle;
					HandleMacCommand();
				} else {
					// Not done, transition to next clock
					mackbd.delay = 18;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
				}
			}
			break;
		case 3:
			// We just entered the read state, bring the clock low
			mackbd.clockstate = 0;
			mackbd.delay = 18;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			break;
		case 4: /* read the data 80uS into the rising clock */
			if(--mackbd.delay == 0) {
				mackbd.cmd |= GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) << (7 - mackbd.bits++);
				mackbd.delay = 14;
				mackbd.clockstate = 1;
			}
			break;
		default:
			break;
		};
		break;
	case kMacKbdStateWriteSettle:
		if(--mackbd.delay == 0) {
			if(GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) == 0) {
				/* "... the computer then indicates that it is ready 
				 * to receive the keyboard's response by setting the 
				 * Keyboard Data line high
				 */
				mackbd.delay = 1;
			} else {
				mackbd.state = kMacKbdStateWriting;
				mackbd.clockstate = 3;
				mackbd.bits = 0;
			}
		}
		break;
	case kMacKbdStateWriting:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.delay = 18;
				mackbd.clockstate = 1;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					mackbd.state = kMacKbdStateIdle;
					GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
					GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
					GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
					GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
					GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
					GPIO_Init(GPIOC, &GPIO_InitStructure);
				} else {
					respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
					mackbd.delay = 16;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
					respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
					mackbd.bits++;
				}
			}
			break;
		case 3:
			GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
			GPIO_Init(GPIOC, &GPIO_InitStructure);

			respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
			mackbd.bits++;
			mackbd.clockstate = 0;
			mackbd.delay = 16;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
			break;
			
		default:
			break;
		};
		break;
	case kMacKbdStateInstantWait:
	case kMacKbdStateInquiryWait:
		if(mackey[keyindex]) {
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = mackey[keyindex];
			mackey[keyindex] = 0;
			keyindex = (keyindex + 1) % kBuffSize;
		} else if(--mackbd.delay == 0) {
			// If no keyboard press, send a null
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = 0x7B;
		}
		break;
	case kMacKbdStateTest:
		mackbd.state = kMacKbdStateWriteSettle;
		mackbd.delay = 1;
		mackbd.response = 0x7D;
		break;
	default:
		break;
	};
}

int main(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	GPIO_InitTypeDef GPIO2_struct;
	int i;

	mousex = 0;
	mousey = 0;
	tenustimer = 0;
	keyindex = 0;
	key2index = 0;
	keypad = 0;
	for(i = 0; i < kBuffSize; i++) mackey[i] = 0;

	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOB, ENABLE);
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);

	// Mouse

	GPIO2_struct.GPIO_Pin = kMSEX1 | kMSEX2 | kMSEY1 | kMSEY2;
	GPIO2_struct.GPIO_Mode = GPIO_Mode_OUT;
	GPIO2_struct.GPIO_OType = GPIO_OType_PP;
	GPIO2_struct.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO2_struct.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO2_struct);

	// Mac Keyboard
	// Clock
	GPIO_InitStructure.GPIO_Pin = kMacKbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Data
	GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	USART2_Init(USART2_Callback);
	InitPS2Mouse();
	InitPS2Keyboard();
	ADBInit();

	// Mouse button
	GPIO_InitStructure.GPIO_Pin = kMSESW;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	USART2_SendString("Starting up\n\r");
	GPIO_SetBits(GPIOC, kMacKbdClockPin);

	mousebits = 0;
	GPIO_SetBits(GPIOC, kMSEX1 | kMSEY1);
	mousebits |= (kMSEX1 | kMSEY1);
	GPIO_ResetBits(GPIOC, kMSEX2 | kMSEY2);
	mousebits &= ~(kMSEX2 | kMSEY2);

	GPIO_SetBits(GPIOA, kMSESW);

	mackbd.state = kMacKbdStateIdle;

	// set a ~10us timer
	SysTick_Config(SystemCoreClock/100000);
	USART2_SendString("Timer configured\n\r");

	SetPS2Repeat();
	SetupPS2Mouse();

	uint8_t ps2status;
	uint8_t ps2xmove;
	uint8_t ps2ymove;

	while(1) {
		int ret;

		if(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0) {
			uint8_t ps2key;
			uint8_t ps2keyup;
			uint8_t ps2extended;

			if(HandlePS2Keyboard(&ps2key, &ps2keyup, &ps2extended)) {
				PS2ToADBScancode(ps2key, ps2keyup, ps2extended);
				PS2ToMacScancode(ps2key, ps2keyup, ps2extended);
			}
		}

		if(GPIO_ReadInputDataBit(GPIOC, kPS2MouseClockPin) == 0) {
			uint8_t absx;
			uint8_t absy;

			ps2status = ps2xmove = ps2ymove = 0;

			ret = HandlePS2Mouse(&ps2status, &ps2xmove, &ps2ymove);
			if(ret == kPS2RetryCode) {
				USART2_SendString("Retrying PS2\n\r");
				continue;
			}
			if(ret != 0) {
				USART2_SendString("PS2 Mouse failure\n\r");
				continue;
			}

			ADBSendMouse(ps2status, ps2xmove, ps2ymove);

			absx = ps2xmove;
			absy = ps2ymove;
			if(ps2xmove && ((mousex & 0x00FF) == 0)) {
				mousex = 0;
				if(ps2status & 0x10) {
					absx = ~ps2xmove;
					absx += 1;
					mousex = 0x0100 | absx | 3;
				} else {
					mousex = absx | 3;
				}
			}
			if(ps2ymove && ((mousey & 0x00FF) == 0)) {
				mousey = 0;
				if(ps2status & 0x20) {
					absy = ~ps2ymove;
					absy += 1;
					mousey = 0x0100 | absy | 3;
				} else {
					mousey = absy | 3;
				}
			}
			if(ps2status & 0x01) {
				GPIO_ResetBits(GPIOA, kMSESW);
			} else {
				GPIO_SetBits(GPIOA, kMSESW);
			}
		}
	}

}

uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended)
{
	uint8_t maccode = 0;
	keypad = 0;
	if(!extended) {
		switch(code) {
		case 0x1C: maccode = 0x01; break; // A
		case 0x32: maccode = 0x17; break; // B
		case 0x21: maccode = 0x11; break; // C
		case 0x23: maccode = 0x05; break; // D
		case 0x24: maccode = 0x1D; break; // E
		case 0x2B: maccode = 0x07; break; // F
		case 0x34: maccode = 0x0B; break; // G
		case 0x33: maccode = 0x09; break; // H
		case 0x43: maccode = 0x45; break; // I
		case 0x3B: maccode = 0x4D; break; // J
		case 0x42: maccode = 0x51; break; // K
		case 0x4B: maccode = 0x4B; break; // L  // woah
		case 0x3A: maccode = 0x5D; break; // M
		case 0x31: maccode = 0x5B; break; // N
		case 0x44: maccode = 0x3F; break; // O
		case 0x4D: maccode = 0x47; break; // P
		case 0x15: maccode = 0x19; break; // Q
		case 0x2D: maccode = 0x1F; break; // R
		case 0x1B: maccode = 0x03; break; // S
		case 0x2C: maccode = 0x23; break; // T
		case 0x3C: maccode = 0x41; break; // U
		case 0x2A: maccode = 0x13; break; // V
		case 0x1D: maccode = 0x1B; break; // W
		case 0x22: maccode = 0x0F; break; // X
		case 0x35: maccode = 0x21; break; // Y
		case 0x1A: maccode = 0x0D; break; // Z
		case 0x45: maccode = 0x3B; break; // 0
		case 0x16: maccode = 0x25; break; // 1
		case 0x1E: maccode = 0x27; break; // 2
		case 0x26: maccode = 0x29; break; // 3
		case 0x25: maccode = 0x2B; break; // 4
		case 0x2E: maccode = 0x2F; break; // 5
		case 0x36: maccode = 0x2D; break; // 6
		case 0x3D: maccode = 0x35; break; // 7
		case 0x3E: maccode = 0x39; break; // 8
		case 0x46: maccode = 0x33; break; // 9
		case 0x0E: maccode = 0x65; break; // `
		case 0x4E: maccode = 0x37; break; // -
		case 0x55: maccode = 0x31; break; // =
		case 0x5D: maccode = 0x55; break; // backslash
		case 0x66: maccode = 0x67; break; // bkspc
		case 0x29: maccode = 0x63; break; // space
		case 0x0D: maccode = 0x61; break; // tab
		case 0x58: maccode = 0x73; break; // caps
		case 0x12: maccode = 0x71; break; // left shift
		case 0x59: maccode = 0x71; break; // right shift
		case 0x11: maccode = 0x75; break; // left alt = opt
		case 0x5A: maccode = 0x49; break; // enter
		case 0x41: maccode = 0x57; break; // ,
		case 0x49: maccode = 0x5F; break; // .
		case 0x4A: maccode = 0x59; break; // /
		case 0x14: maccode = 0x77; break; // L control
		case 0x76: maccode = 0x6B; break; // ESC
		case 0x52: maccode = 0x4F; break; // '
		case 0x7C: maccode = 0x79; keypad = 0x05; break; // keypad *
		case 0x7B: maccode = 0x79; keypad = 0x1D; break; // keypad -
		case 0x79: maccode = 0x79; keypad = 0x0D; break; // keypad +
		case 0x71: maccode = 0x79; keypad = 0x03; break; // keypad .
		case 0x70: maccode = 0x79; keypad = 0x25; break; // keypad 0
		case 0x69: maccode = 0x79; keypad = 0x27; break; // keypad 1
		case 0x72: maccode = 0x79; keypad = 0x29; break; // keypad 2
		case 0x7A: maccode = 0x79; keypad = 0x2B; break; // keypad 3
		case 0x6B: maccode = 0x79; keypad = 0x2D; break; // keypad 4
		case 0x73: maccode = 0x79; keypad = 0x2F; break; // keypad 5
		case 0x74: maccode = 0x79; keypad = 0x31; break; // keypad 6
		case 0x6C: maccode = 0x79; keypad = 0x33; break; // keypad 7
		case 0x75: maccode = 0x79; keypad = 0x37; break; // keypad 8
		case 0x7D: maccode = 0x79; keypad = 0x39; break; // keypad 9
		default: break;
		}
	} else {
		switch(code) {
		case 0x1F: maccode = 0x6F; break; // left windows key = cmd
		case 0x27: maccode = 0x6F; break; // right windows key = cmd
		case 0x11: maccode = 0x75; break; // right alt = cmd
		case 0x5A: maccode = 0x79; keypad = 0x19; break; // keypad enter
		case 0x4A: maccode = 0x79; keypad = 0x1B; break;  // keypad /
		case 0x75: maccode = 0x79; keypad = 0x1B; break;  // up -> keypad up
		case 0x6B: maccode = 0x79; keypad = 0x0D; break;  // left arrow
		case 0x72: maccode = 0x79; keypad = 0x11; break;  // down arrow
		case 0x74: maccode = 0x79; keypad = 0x05; break;  // right arrow
		case 0x14: maccode = 0x77; break; // R control
		}
	}

	if(keyup && maccode) {
		if(keypad) {
			keypad |= 0x80;
		} else {
			maccode |= 0x80;
		}
	}

	if(maccode == 0) maccode = 0x7B;

	__disable_irq();
	if(((key2index+1)%kBuffSize) != keyindex) {
		mackey[key2index] = maccode;
		key2index = (key2index+1) % kBuffSize;
		if(keypad) {
			mackey[key2index] = keypad;
			key2index = (key2index+1) % kBuffSize;
		}
	}
	__enable_irq();

	return maccode;
}

void HandleMacCommand()
{
	switch(mackbd.cmd) {
	case 0x16:
		mackbd.state = kMacKbdStateWriteSettle;
		mackbd.delay = 10;
		mackbd.response = 0x03;
		break;
	case 0x14:
		mackbd.state = kMacKbdStateInstantWait;
		mackbd.delay = 1;
		break;
	case 0x36:
		mackbd.state = kMacKbdStateTest;
		mackbd.delay = 10;
		break;
	case 0x10:
	default:
		mackbd.state = kMacKbdStateInquiryWait;
		mackbd.delay = 25000;
		break;
	};

	return;
}

uint8_t PS2ToADBScancode(uint8_t code, uint8_t keyup, uint8_t extended)
{
	uint8_t maccode = 0xFF;
	if(!extended) {
		switch(code) {
		case 0x1C: maccode = 0x00; break; // A
		case 0x32: maccode = 0x0B; break; // B
		case 0x21: maccode = 0x08; break; // C
		case 0x23: maccode = 0x02; break; // D
		case 0x24: maccode = 0x0E; break; // E
		case 0x2B: maccode = 0x03; break; // F
		case 0x34: maccode = 0x05; break; // G
		case 0x33: maccode = 0x04; break; // H
		case 0x43: maccode = 0x22; break; // I
		case 0x3B: maccode = 0x26; break; // J
		case 0x42: maccode = 0x28; break; // K
		case 0x4B: maccode = 0x25; break; // L
		case 0x3A: maccode = 0x2E; break; // M
		case 0x31: maccode = 0x2D; break; // N
		case 0x44: maccode = 0x1F; break; // O
		case 0x4D: maccode = 0x23; break; // P
		case 0x15: maccode = 0x0C; break; // Q
		case 0x2D: maccode = 0x0F; break; // R
		case 0x1B: maccode = 0x01; break; // S
		case 0x2C: maccode = 0x11; break; // T
		case 0x3C: maccode = 0x20; break; // U
		case 0x2A: maccode = 0x09; break; // V
		case 0x1D: maccode = 0x0D; break; // W
		case 0x22: maccode = 0x07; break; // X
		case 0x35: maccode = 0x10; break; // Y
		case 0x1A: maccode = 0x06; break; // Z
		case 0x45: maccode = 0x1D ;break; // 0
		case 0x16: maccode = 0x12; break; // 1
		case 0x1E: maccode = 0x13; break; // 2
		case 0x26: maccode = 0x14; break; // 3
		case 0x25: maccode = 0x15; break; // 4
		case 0x2E: maccode = 0x17; break; // 5
		case 0x36: maccode = 0x16; break; // 6
		case 0x3D: maccode = 0x1A; break; // 7
		case 0x3E: maccode = 0x1C; break; // 8
		case 0x46: maccode = 0x19; break; // 9
		case 0x0E: maccode = 0x32; break; // `
		case 0x4E: maccode = 0x1B; break; // -
		case 0x55: maccode = 0x18; break; // =
		case 0x5D: maccode = 0x2A; break; // backslash
		case 0x66: maccode = 0x33; break; // bkspc
		case 0x29: maccode = 0x31; break; // space
		case 0x0D: maccode = 0x30; break; // tab
		case 0x58: maccode = 0x39; break; // caps
		case 0x12: maccode = 0x38; break; // left shift
		case 0x59: maccode = 0x7B; break; // right shift
		case 0x11: maccode = 0x3A; break; // left alt = opt
		case 0x5A: maccode = 0x24; break; // enter
		case 0x41: maccode = 0x2B; break; // ,
		case 0x49: maccode = 0x2F; break; // .
		case 0x4A: maccode = 0x2C; break; // /
		case 0x14: maccode = 0x36; break; // L control
		case 0x76: maccode = 0x35; break; // ESC
		case 0x52: maccode = 0x27; break; // '
		case 0x4C: maccode = 0x29; break; // ;
		case 0x7C: maccode = 0x43; break; // keypad *
		case 0x7B: maccode = 0x4E; break; // keypad -
		case 0x79: maccode = 0x45; break; // keypad +
		case 0x71: maccode = 0x41; break; // keypad .
		case 0x70: maccode = 0x52; break; // keypad 0
		case 0x69: maccode = 0x53; break; // keypad 1
		case 0x72: maccode = 0x54; break; // keypad 2
		case 0x7A: maccode = 0x55; break; // keypad 3
		case 0x6B: maccode = 0x56; break; // keypad 4
		case 0x73: maccode = 0x57; break; // keypad 5
		case 0x74: maccode = 0x58; break; // keypad 6
		case 0x6C: maccode = 0x59; break; // keypad 7
		case 0x75: maccode = 0x5B; break; // keypad 8
		case 0x7D: maccode = 0x5C; break; // keypad 9
		case 0x05: maccode = 0x7A; break; // F1
		case 0x06: maccode = 0x78; break; // F2
		case 0x04: maccode = 0x63; break; // F3
		case 0x0C: maccode = 0x76; break; // F4
		case 0x03: maccode = 0x60; break; // F5
		case 0x0B: maccode = 0x61; break; // F6
		case 0x83: maccode = 0x62; break; // F7
		case 0x0A: maccode = 0x64; break; // F8
		case 0x01: maccode = 0x65; break; // F9
		case 0x09: maccode = 0x6D; break; // F10
		case 0x78: maccode = 0x67; break; // F11
		case 0x07: maccode = 0x6F; break; // F12
		default: break;
		}
	} else {
		switch(code) {
		case 0x1F: maccode = 0x37; break; // left windows key = cmd
		case 0x27: maccode = 0x37; break; // right windows key = cmd
		case 0x11: maccode = 0x7C; break; // right alt = cmd
		case 0x5A: maccode = 0x4C; break; // keypad enter
		case 0x4A: maccode = 0x4B; break;  // keypad /
		case 0x75: maccode = 0x3E; break;  // up -> keypad up
		case 0x6B: maccode = 0x3B; break;  // left arrow
		case 0x72: maccode = 0x3D; break;  // down arrow
		case 0x74: maccode = 0x3C; break;  // right arrow
		case 0x14: maccode = 0x7D; break; // R control
		}
	}

#define MODKEYSET(y) keyup ? (modifierkeys |= (y)) : (modifierkeys &= ~(y))
	switch(maccode) {
	case 0x37: MODKEYSET(kModCmd); break; 
	case 0x7C:
	case 0x3A: MODKEYSET(kModOpt); break; 
	case 0x7B:
	case 0x38: MODKEYSET(kModShift); break; 
	case 0x7D:
	case 0x36: MODKEYSET(kModControl); break;
	case 0x39: MODKEYSET(kModCaps); break;
	case 0x33: MODKEYSET(kModDelete); break;
	}

	if((!(modifierkeys & kModShift)) && (maccode == 0x35)) {
		MODKEYSET(kModReset);
		if(keyup) {
			GPIO_SetBits(GPIOB, kADBPowerPin);
			USART2_SendString("Power key up\n\r");
			__disable_irq();
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0xFF;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x38;
			adbend = (adbend+1) % kADBBuffSize;
			adb.senddata |= (1<<2);
			__enable_irq();
		} else {
			GPIO_ResetBits(GPIOB, kADBPowerPin);
			USART2_SendString("Power key down\n\r");
			__disable_irq();
			adbkey[adbend] = 0x38 | 0x80;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adbkey[adbend] = 0x7F;
			adbend = (adbend+1) % kADBBuffSize;
			adb.senddata |= (1<<2);
			__enable_irq();
		}
		return 0x7F;
	}

	if(keyup) {
		maccode |= 0x80;
	}

	__disable_irq();
	adbkey[adbend] = maccode;
	adbend = (adbend+1) % kBuffSize;
	adb.senddata |= (1<<2);
	__enable_irq();

	return maccode;
}

void USART2_Callback(uint8_t c)
{
	switch(c) {
	case 'r': 
		mousex = 100;
		adb.device[3].reg0 = 0x8000 | (10 & 0x7F) | 0x80;
		adb.senddata |= (1<<3);
		break;
	case 'l': 
		mousex = -100;
		adb.device[3].reg0 = 0x8000 | ((-10) & 0x7F) | 0x80;
		adb.senddata |= (1<<3);
		break;
	case 'u': 
		mousey = -100;
		adb.device[3].reg0 = 0x8000 | (((-10) & 0x7F) << 8) | 0x80;
		adb.senddata |= (1<<3);
		break;
	case 'd': 
		mousey = 100;
		adb.device[3].reg0 = 0x8000 | ((10 & 0x7F) << 8) | 0x80;
		adb.senddata |= (1<<3);
		break;
	}
}
