#include "stm32f0xx_conf.h"
#include "ps2.h"
#include "usart.h"

extern __IO uint32_t tenustimer;
uint8_t ps2ledstate = 0;

void InitPS2Mouse(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	GPIO_InitStructure.GPIO_Pin = kPS2MouseDataPin | kPS2MouseClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);
}

void InitPS2Keyboard(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	GPIO_InitStructure.GPIO_Pin = kPS2KbdDataPin | kPS2KbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);
}

int HandlePS2Mouse(uint8_t *status, uint8_t *xmove, uint8_t *ymove)
{
	uint16_t ps2frame = 0;

#if 0
	// Wait for any current transactions to be over.
	hightime = tenustimer;
	while((tenustimer - hightime) < 1000) {
		if(GPIO_ReadInputDataBit(GPIOC, kPS2MouseClockPin) == 0) {
			hightime = tenustimer;
		}
	}
#endif
	while(GPIO_ReadInputDataBit(GPIOC, kPS2MouseClockPin) == 1);

	ps2frame = 0;
	do {
		ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
		if(ps2frame == kPS2ErrorCode) {
			USART2_SendString("Failure on status\n\r");
			return -1;
		}
	}while(ps2frame == kPS2RetryCode);
	*status = (ps2frame >> 1) & 0xFF;
	if(*status == 0) {
		GPIO_SetBits(GPIOC, GPIO_Pin_5);
	}

	// self-test passed, a mouse was probably just attached
	if(*status == 0xAA) {
		// Read mouse id
		ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
		
		SetupPS2Mouse();
		USART2_SendString("Re-setup mouse\n\r");
		return -1;
	}

	while(GPIO_ReadInputDataBit(GPIOC, kPS2MouseClockPin) == 1);
	ps2frame = 0;
	do {
		ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
		if(ps2frame == kPS2ErrorCode) {
			USART2_SendString("Failure on xmove\n\r");
			return -1;
		}
	}while(ps2frame == kPS2RetryCode);
	*xmove = (ps2frame >> 1) & 0xFF;

	while(GPIO_ReadInputDataBit(GPIOC, kPS2MouseClockPin) == 1);
	ps2frame = 0;
	do {
		ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
		if(ps2frame == kPS2ErrorCode) {
			USART2_SendString("Failure on ymove\n\r");
			return -1;
		}
	}while(ps2frame == kPS2RetryCode);
	*ymove = (ps2frame >> 1) & 0xFF;

	if(!(*status & 0x08)) {
		USART2_SendString("Status is missing bit 0x08: ");
		USART2_SendHexByte(*status);
		USART2_SendString("\n\r");
		return -1;
	}

#if 0
	USART2_SendString("PS2: ");
	USART2_SendHexByte(*status);
	USART2_SendHexByte(*xmove);
	USART2_SendHexByte(*ymove);
	USART2_SendString("\n\r");
#endif

	return 0;
}

uint8_t HandlePS2Keyboard(uint8_t *code, uint8_t *keyup, uint8_t *extended) 
{
	static uint8_t ps2lastcode = 0;
	uint16_t ps2frame = 0;
	uint8_t ps2scancode;
	uint8_t tmpcode = 0;

	*code = *keyup = *extended = 0;

	ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
	ps2scancode = tmpcode = (ps2frame >> 1) & 0xFF;
	if(ps2scancode == 0xF0) {
		// Keyup, poll in the scancode of the key that has been released
		ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
		ps2scancode = ps2frame >> 1;
		// Ignore keyup capslock if we're turning it on
		if(((ps2scancode) == 0x58) && (ps2ledstate & kPS2LEDCaps)) {
		} else {
			*code = ps2scancode;
			*keyup = 1;
			*extended = 0;
		}
	} else if(ps2scancode == 0xE0) {
		// two byte scan code
		ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
		ps2scancode = ps2frame >> 1;
		if(ps2scancode == 0xF0) {
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			if(ps2scancode == 0x7C) {
				// prntscrn
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be E0
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be F0
				ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
				ps2scancode = ps2frame >> 1;
				// should be 12
				return 0;
			}
			*code = ps2scancode;
			*keyup = 1;
			*extended = 1;
		} else if(ps2scancode == 0x12) {
			// prntscrn
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			// should be E0
			ps2frame = PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
			ps2scancode = ps2frame >> 1;
			// should be 7C

			return 0;
		} else {
			*code = ps2scancode;
			*keyup = 0;
			*extended = 1;
		}
	} else {
		if(ps2scancode == ps2lastcode) {
			return 0;
		}
		// If it's capslock, turn on the capslock led
		if((ps2scancode) == 0x58) {
			if(ps2ledstate & kPS2LEDCaps) {
				ps2ledstate &= ~kPS2LEDCaps;
			} else {
				ps2ledstate |= kPS2LEDCaps;
			}
			SetPS2LEDState(ps2ledstate);
		}
		*code = ps2scancode;
		*keyup = 0;
		*extended = 0;
	}
	ps2lastcode = tmpcode;
	return ps2scancode;
}

void SetupPS2Mouse(void)
{
	uint16_t ps2frame = 0;
	uint8_t ps2scancode;
	uint8_t resetresult;
	uint8_t mouseid;
	uint8_t result;
	unsigned long tmp1 = tenustimer;

	while(tenustimer < (tmp1+10));
	
	USART2_SendString("Initializing mouse\n\r");
	USART2_SendString("Sending mouse reset: ");
	// reset
	SendPS2Command(0xFF, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	result = ps2frame >> 1;
	
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	resetresult = ps2frame >> 1;

	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	mouseid = ps2frame >> 1;

	USART2_SendHexByte(result);
	USART2_SendString(" ");
	USART2_SendHexByte(resetresult);
	USART2_SendString(" ");
	USART2_SendHexByte(mouseid);
	USART2_SendString("\n\r");

	USART2_SendString("Setting sample rate: ");
	// Set sample rate
	SendPS2Command(0xF3, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	result = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x64, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	USART2_SendHexByte(result);
	USART2_SendString(" ");
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");

	USART2_SendString("reading device type: ");
	// Read device type
	SendPS2Command(0xF2, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	result = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be mouseid
	USART2_SendHexByte(result);
	USART2_SendString(" ");
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");

	USART2_SendString("Setting resolution: ");
	// Set resolution
	SendPS2Command(0xE8, kPS2MouseClockPin, kPS2MouseDataPin);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2MouseDataPin) == 1);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	result = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x04, kPS2MouseClockPin, kPS2MouseDataPin);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2MouseDataPin) == 1);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	USART2_SendHexByte(result);
	USART2_SendString(" ");
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");

	USART2_SendString("Setting scaling: ");
	// Set scaling
	SendPS2Command(0xE6, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");

#if 0
	USART2_SendString("Setting sample rate: ");
	// Set sample rate
	SendPS2Command(0xE6, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	result = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	SendPS2Command(0x28, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	USART2_SendHexByte(result);
	USART2_SendString(" ");
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");
#endif
	

	USART2_SendString("Enabling: ");
	// Enable
	SendPS2Command(0xF4, kPS2MouseClockPin, kPS2MouseDataPin);
	ps2frame = PollPS2Frame(kPS2MouseClockPin, kPS2MouseDataPin);
	ps2scancode = ps2frame >> 1;
	// ps2scancode should be 0xFA for ack
	USART2_SendHexByte(ps2scancode);
	USART2_SendString("\n\r");
}

void SetPS2LEDState(uint8_t cmd)
{
	unsigned long tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(0xED, kPS2KbdClockPin, kPS2KbdDataPin);
	// wait for ack 0xFA
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
	tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(cmd, kPS2KbdClockPin, kPS2KbdDataPin);
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
}

void SetPS2Repeat(void)
{
	unsigned long tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(0xF8, kPS2KbdClockPin, kPS2KbdDataPin);
	// wait for ack 0xFA
	PollPS2Frame(kPS2KbdClockPin, kPS2KbdDataPin);
}

void SendPS2Command(uint8_t cmd, uint32_t clock, uint32_t data)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	unsigned long tmp1;
	uint8_t bits;
	uint8_t parity = 0;
	uint32_t timeout;

	GPIO_InitStructure.GPIO_Pin = clock;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	tmp1 = tenustimer;
	GPIO_ResetBits(GPIOC, clock);
	// pull clock low for >=100us
	while(tenustimer < (tmp1 + 15));

	// Bring data line low
	GPIO_InitStructure.GPIO_Pin = data;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);
	GPIO_ResetBits(GPIOC, data);
	
	// Release clock
	GPIO_InitStructure.GPIO_Pin = clock;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Might need to wait for the clock to go high after
	// releasing it...
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}

#if 0
	// start bit
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1);
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0);
#endif

	for(bits = 0; bits < 8; bits++) {
		// Wait for device to bring clock low
		timeout = tenustimer;
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 1) {
			if((tenustimer - timeout) >= kPollTimeout) {
				return;
			}
		}
		if(cmd & (1<<bits)) {
			GPIO_SetBits(GPIOC, data);
			parity++;
		} else {
			GPIO_ResetBits(GPIOC, data);
		}
		// wait out the remainder of the low clock
		timeout = tenustimer;
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 0) {
			if((tenustimer - timeout) >= kPollTimeout) {
				return;
			}
		}
	}

	// Data has been sent, now send parity
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}
	if((parity & 0x01) == 0) {
		GPIO_SetBits(GPIOC, data);
	} else {
		GPIO_ResetBits(GPIOC, data);
	}
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 0) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}

	// Stop bit, release data line
	GPIO_InitStructure.GPIO_Pin = data;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, data) == 0) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, data) == 1) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}
	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, data) == 0) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return;
		}
	}
}

uint16_t PollPS2Frame(uint32_t clock, uint32_t data)
{
	uint16_t ps2frame = 0;
	uint8_t bits;
	uint32_t timeout;
	uint8_t parity = 0;

	GPIO_SetBits(GPIOC, GPIO_Pin_5);

	timeout = tenustimer;
	while(GPIO_ReadInputDataBit(GPIOC, clock) == 1) {
		if((tenustimer - timeout) >= kPollTimeout) {
			return kPS2ErrorCode;
		}
	}

	for(bits = 0; bits < 11; bits++) {

		// device to host should be read on falling edge
		ps2frame |= GPIO_ReadInputDataBit(GPIOC, data) << bits;

		timeout = tenustimer;
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 0) {
			if((tenustimer - timeout) >= kPollTimeout) {
				GPIO_ResetBits(GPIOC, GPIO_Pin_5);
				USART2_SendString("PS2 Low Timeout\n\r");
				return kPS2ErrorCode;
			}
		}
		if(bits == 10) break;

		timeout = tenustimer;
		while(GPIO_ReadInputDataBit(GPIOC, clock) == 1) {
			if((tenustimer - timeout) >= kPollTimeout) {
				GPIO_ResetBits(GPIOC, GPIO_Pin_5);
				USART2_SendString("PS2 High Timeout\n\r");
				return kPS2ErrorCode;
			}
		}
	}

	if(ps2frame & 1) {
		GPIO_ResetBits(GPIOC, GPIO_Pin_5);
		USART2_SendString("Start bit failure on ");
		USART2_SendHexByte(ps2frame >> 8);
		USART2_SendHexByte(ps2frame);
		USART2_SendString("\n\r");
		return kPS2ErrorCode;
	}
	if(!(ps2frame & (1<<10))) {
		GPIO_ResetBits(GPIOC, GPIO_Pin_5);
		USART2_SendString("Stop bit failure on ");
		USART2_SendHexByte(ps2frame >> 8);
		USART2_SendHexByte(ps2frame);
		USART2_SendString("\n\r");
		return kPS2ErrorCode;
	}

	// Enforce parity
	for(bits = 1; bits < 9; bits++) {
		if(ps2frame & (1<<bits)) parity++;
	}
	if((ps2frame & (1<<9)) && ((parity & 0x01) == 0)) return ps2frame;
	if(!(ps2frame & (1<<9)) && ((parity & 0x01) == 1)) return ps2frame;

	GPIO_ResetBits(GPIOC, GPIO_Pin_5);
	USART2_SendString("Parity failure on ");
	USART2_SendHexByte(ps2frame >> 8);
	USART2_SendHexByte(ps2frame);
	USART2_SendString("\n\r");
	return kPS2ErrorCode;
}
